#!/usr/bin/python

import getopt
import yaml
import sys
import signal

from engine import GsyncEngine
from utils.daemon import Daemon
from utils.logger import log
from repository.remote import HTTPRemoteConnection

DEFAULTS={
'RunStandalone': True,
'Verbose': True,
'LogFile': '',
'Debug': '',
'ConfigFile': '/etc/gandi/gsync.conf',
}
CONFIG = DEFAULTS.copy()

class ConfigError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class GsyncEngined(Daemon):
    def __init__(self):
        self.config = CONFIG
        self.ConfigFile = CONFIG['ConfigFile']
        self.gsync_config = None

    def commandLine(self):
        try:
            options, args = getopt.getopt(sys.argv[1:], "hvf:XT:p")
        except:
            self.usage()
            sys.exit(1)

        # Parse all the options
        for option, value in options:
            if option == "-h":
                self.usage()
                return
            elif option == "-f":
                self.ConfigFile = value
            elif option == "-v":
                self.config['Verbose'] = True
            elif option == "-X":
                self.config['RunStandalone'] = False
            else:
                print "Unrecognized option: %s" % (option,)
                self.usage()
                raise ValueError

        self.parseConfig()

    def parseConfig(self):
        try:
            fd = open(self.ConfigFile, 'r')
            conf = yaml.load(fd.read())
            conf_keys = conf.keys()
            t = [x for x in conf_keys if x not in ['webserver', 
                'loglevel', 'repositories']]
            if t:
                raise ConfigError('Unexpected key %s at conf top-level' % t[0])
            if not 'repositories' in conf_keys:
                raise ConfigError('repositories section is missing')
            #Testing remote repository paths 
            remote_repo_paths = [conf['repositories'][x]['path'] for x in conf['repositories']
                if conf['repositories'][x]['type'] == 'remote']
            for path in remote_repo_paths:
                try:
                    conn = HTTPRemoteConnection(path)
                    conn.clone().head()
                except :
                    raise Exception('repository with path %s seems unvailable' % path)
            self.gsync_config = conf
            fd.close()
        except ConfigError, e:
            raise  
        except Exception, e:
            print "Unvalid Syntax in configuration file %s:" % self.ConfigFile
            raise

    def stop(self, sig, stack):
        self.engine.stop()
        sys.exit(0)

    def main(self):
        signal.signal(signal.SIGTERM, self.stop)
        self.engine = GsyncEngine()
        if self.engine.setup(self.gsync_config):
            self.engine.run()

    def start(self):
        if self.config['RunStandalone']:
            self.daemonize()
        else:
            self.main()

    def usage(self):
        print """Usage: gsyncd [options] start|stop|restart|debug
Options:
    -h          Print this help and exit
    -v          Be verbose
    -f config   Specify path to configuration file [""" + CONFIG['ConfigFile'] + """]
    -X          Do not daemonize
    -p          Print current configuration and exit
"""

if __name__ == "__main__":
    try:
        daemon = GsyncEngined()
        daemon.commandLine()
        daemon.start()
    except (KeyboardInterrupt, Exception), e:
        log.err('exiting: %s:%s' % (e.__class__, e))
        sys.exit(1)
