import threading
import web.webapi as web
import web.utils as utils
from web.wsgiserver import CherryPyWSGIServer
from BaseHTTPServer import BaseHTTPRequestHandler

from utils.logger import log

class ThreadedCherryPyWSGIServer(CherryPyWSGIServer, threading.Thread):
    def __init__(self, *args, **kwargs):
        CherryPyWSGIServer.__init__(self, *args, **kwargs)
        threading.Thread.__init__(self)
        web.debug.write = log.debug

    def start(self, *args):
        threading.Thread.start(self, *args)

    def run(self, *args):
        CherryPyWSGIServer.start(self, *args)

def runsimple(func, conf):
    """
    Runs [CherryPy][cp] WSGI server hosting WSGI app `func`. 
    [cp]: http://www.cherrypy.org
    """
    class WSGIWrapper(BaseHTTPRequestHandler):
        """WSGI wrapper for logging the status and serving static files."""
        def __init__(self, app):
            self.app = app
            self.format = '%s - - [%s] "%s %s %s" - %s'

        def __call__(self, environ, start_response):
            def xstart_response(status, response_headers, *args):
                write = start_response(status, response_headers, *args)
                self.log(status, environ)
                return write

            path = environ.get('PATH_INFO', '')
            return self.app(environ, xstart_response)

        def log(self, status, environ):
            outfile = environ.get('wsgi.errors', web.debug)
            req = environ.get('PATH_INFO', '_')
            protocol = environ.get('ACTUAL_SERVER_PROTOCOL', '-')
            method = environ.get('REQUEST_METHOD', '-')
            host = "%s:%s" % (environ.get('REMOTE_ADDR','-'),
                              environ.get('REMOTE_PORT','-'))

            #@@ It is really bad to extend from 
            #@@ BaseHTTPRequestHandler just for this method
            time = self.log_date_time_string()

            msg = self.format % (host, time, protocol, method, req, status)
            log.debug(msg)

    server_address = ("0.0.0.0", conf.get('port', 8080))
    func = WSGIWrapper(func)
    server = ThreadedCherryPyWSGIServer(server_address, func, server_name="localhost")
    if conf.get('secure', False):
        server.ssl_certificate = conf['secure']['ssl_certificate']
        server.ssl_private_key = conf['secure']['ssl_private_key']

    log.debug('starting webserver on http://%s:%d/' % server_address)
    return server
