"""
Basic logging function
"""

import posixpath
import syslog
import traceback
import sys

class Logger:
    levels = {
        'err': syslog.LOG_ERR,
        'warn': syslog.LOG_WARNING,
        'info': syslog.LOG_INFO,
        'debug': syslog.LOG_DEBUG
    }

    def __init__(self, name, level=syslog.LOG_WARNING):
        if isinstance(level, str):
            level = self.levels[level]

        self.level = level
        self.name = name
        syslog.openlog(name, syslog.LOG_PID, syslog.LOG_LOCAL7)

    def set_level(self, level):
        """Change our log level"""
        self.level = self.levels.get(level, 'info')
            
    def debug(self, msg):
        if self.level >= self.levels['debug']:
            self.do_log(msg, 'debug')

    def info(self, msg):
        if self.level >= self.levels['info']:
            self.do_log(msg, 'info')

    def warn(self, msg):
        if self.level >= self.levels['warn']:
            self.do_log(msg, 'warn')

    def err(self, msg):
        self.do_log(msg, 'err')
        if self.level >= self.levels['debug']:
            _, e, tb = sys.exc_info()
            if tb:
                for st in self.printable_stacks(tb):
                    msg = ' TB -> %s' % st
                    self.do_log(msg, 'err')
            else:
                self.do_log('(no traceback)', 'err')

    def printable_stacks(self, tb=None):
        st = []
        tbl = traceback.extract_stack()[:-3]
        if tb:
            tbl = traceback.extract_tb(tb)

        for (file, lnr, func, line) in tbl:
            file = '/'.join(file.split('/')[-2:])
            st.append('%s:%s' % (file, lnr,))

        return st

    def do_log(self, msg, lev):
        if self.level >= self.levels['debug']:
            prefix = self.printable_stacks()[-1]
        else:
            prefix = lev.upper()

        msg = '<%s> %s' % (prefix, msg)

        if sys.stdout.isatty():
            print msg
        else:
            syslog.syslog(self.levels[lev], msg)

_default_logger = Logger(sys.argv[0], Logger.warn)
log = _default_logger
