#!/usr/bin/env python
import web
import os
import stat
import cgi
import re
#import json
from datetime import datetime
from cStringIO import StringIO
from utils.mimerender import mimerender
from mimetypes import guess_type
from base64 import b64decode
import urllib

from utils.logger import log

HTMLTEMPLATE = """
<title>Directory listing for $displaypath - $repository_name</title>\n
<link rel="StyleSheet" href="/static/styles/directory_listing.css" type="text/css" media="screen">
<link rel="shortcut icon" href="/static/images/favicon.ico"/>
<div id="page">
    <div id="header">
        <h2>Directory listing for $displaypath  - $repository_name</h2>\n
    </div>
    <div id="content" class="widecolumn" style="width:700px">
    <div class="entry files post">
    <pre>
    <table style="width:100%;text-align:justify;">
    <tr><th style="width:60%">File Name</th><th style="width:10%">Size</th><th style="width:30%">Modified Date</th></tr>
    <tr><td colspan="3"><hr/></td></tr>
    <tr><td colspan="3" style="padding-bottom:20px"><a href="/"><img width="15" height="11" alt="[DIR]" src="/static/images/back.png"/> <a href="/repositories/$repository_name/$path/..">Parent folder</a></td></tr>

    $for i in files:
        <tr><td><a href="$i['linkname']">$i['displayname']</a></td><td>$i['size']</td><td>$i['mtime']</td></tr>

    </table>
    </pre>
    </div>
    </div>
    <div id="footer">
        <a href="/">Repositories Index</a>\n
    </div>
</div>
""" 
XMLTEMPLATE = """
<files>
$for i in files:
    <file>$i['linkname']</file>
</files>
"""
#render_xml = lambda **vars: web.template.Template(XMLTEMPLATE, globals=vars)()
#render_json = lambda **args: json.dumps(args)
render_html = lambda **vars: web.template.Template(HTMLTEMPLATE, globals=vars)()
render_txt = lambda **vars: '\n'.join([i['name'] for i in vars['files']])

realm = "Gsync protected access"

class GsyncWebserver(object):
    def __init__(self):
        self.urls = [
          '/static/(.*)', 'Static', 
          '/', 'Index',
        ]
        
        self.gsync_urls = []
        self.repositories = {}
        self.accesses = {}
        self.app = None

    def add_repository(self, repo, access=None):
        repo_name = repo.name
        self.repositories[repo_name] = repo
        if access:
            self.accesses[repo_name] = access
            log.debug('added %s to webserver' % repo)
        self.gsync_urls.extend(['/(%s)/(.*)/?' % repo_name, 'Gsync'])

    def setup(self):
        globals()['accesses'] = self.accesses
        globals()['repositories'] = self.repositories

        gsync_urls = tuple(self.gsync_urls)
        auth_app = web.application(gsync_urls, globals())
        auth_app.add_processor(middleware(auth_mw))

        urls = ['/repositories', auth_app]
        urls.extend(self.urls)
        urls = tuple(urls)
        app = web.application(urls, globals())
        app.notfound = notfound
        app.internalerror = internalerror
        self.app = app

class Index:
    def GET(self, path=''):
        keys = repositories.keys()
        if keys:
            keys.sort(key=lambda a: a.lower())
            f = StringIO()
            f.write("<title>Repositories</title>\n")
            f.write('<link rel="StyleSheet" href="/static/styles/directory_listing.css" type="text/css" media="screen">\n')
            f.write('<link rel="shorcut icon" href="/static/images/favicon.ico"/>')
            f.write('<div id="page"><div id="header"><h2>Repositories</h2></div><div id="content">')
            f.write("<hr>\n<ul>\n")
            for name in keys:
                linkname = 'repositories/' + name
                displayname = name
                f.write('<li><a href="/%s/">%s</a>\n'
                        % (urllib.quote(linkname), cgi.escape(displayname)))
            f.write("</ul>\n<hr>\n")
            f.write('</div></div>')
            length = f.tell()
            f.seek(0)

            web.header("Content-type", "text/html")
            web.header("Content-Length", str(length))
            return  f.read()

class Logout:
    def GET(self):
        #There is no safe way to logout from HTTP Authorization 
        #as HTTP/1.1. does not provide a method for a server to direct clients 
        #to discard the cached credentials
        web.ctx.status = "401 UNAUTHORIZED"
        web.header("WWW-Authenticate", 'Basic realm="%s"'  % web.websafe(realm) )
        return  defaultdeny()

class Static:
    def GET(self, path):
        raise web.seeother('http://open.gandi.net/static/gsync/' + path)

class Gsync:
    def HEAD(self, repository_name, path=''):
        environ = web.webapi.ctx.environ
        metas = web.input()
        if repository_name not in repositories:
            return web.notfound()
        repository = repositories[repository_name]
        ver = metas.get('version', None)
        if ver == 'None': ver = None
        if repository.has(path):
            st = repository.stat(path, ver)
            atts = ['mode', 'ino', 'dev', 'nlink', 'uid',
                    'gid', 'size', 'atime', 'mtime', 'ctime']
            headers = [ ('x-gsync-stat-' + i, 
                str(st[atts.index(i)])) for i in atts ]
            for header, value in headers:
                web.header(header, value)
            if metas:
                try:
                    for (func_name, func_arg) in metas.items():
                        if not func_arg:
                            ret = getattr(repository,func_name)()
                            
                            if ret and type(ret) == type([]):
                                ret = ','.join([str(i)for i in ret])
                                web.header('x-gsync-' + func_name, ret)
                            elif ret and type(ret) == type({}):
                                for k, v in ret.items():
                                    web.header('x-gsync-%s-%s' % (func_name, k), v)
                            else:
                                web.header('x-gsync-' + func_name, ret)
                except Exception, e:
                    log.err("in HEAD %s:%s" % (e.__class__, e))
                    return web.internalerror()
            return 'done'
        else:
            return web.notfound()

    def PUT(self, repository_name, path=''):
        # Maximum input
        # 0 ==> unlimited input
        cgi.maxlen = 1024 * 1024 * 1024
        environ = web.webapi.ctx.environ
        web.webapi.ctx['data'] = ''
        metas = web.input(metas={})
        atts = ['mode', 'ino', 'dev', 'nlink', 'uid', 
                'gid', 'size', 'atime', 'mtime', 'ctime']
        for a in atts:
            if metas.has_key(a):
                try:
                    metas[a] = int(metas[a])
                except:
                    metas[a] = float(metas[a])

        if repository_name not in repositories:
            log.err("Repository %s doen't exist" % repository_name)
            web.internalerror()
            return 'done'
        repository = repositories[repository_name]
        try:
            mode = metas.get('mode')
            if not mode:
                web.badrequest()

            try:
                if stat.S_ISDIR(mode):
                    repository.mkdir(path, metas)
                elif stat.S_ISLNK(mode):
                    if not metas.get('symlink'):
                        web.badrequest()
                    repository.symlink(path, metas['symlink'])
                elif stat.S_ISREG(mode):
                    fd = environ['wsgi.input']
                    content_length = environ.get('CONTENT_LENGTH')
                    if content_length: content_length = int(content_length)
                    repository.add_from_fd(path, fd, metas, content_length)
            except (OSError, IOError), e:
                return web.notfound()
        except Exception, e:

            log.err("in PUT %s:%s" % (e.__class__, e))
            web.internalerror()
        return 'done'
        
    def POST(self, repository_name, path=''):
        environ = web.webapi.ctx.environ
        if repository_name not in repositories:
            return web.notfound()

        repository = repositories[repository_name]
        web.webapi.ctx['data'] = ''
        metas = web.input(metas={})
        atts = ['mode', 'ino', 'dev', 'nlink', 'uid', 
                'gid', 'size', 'atime', 'mtime', 'ctime']
        for a in atts:
            if metas.has_key(a):
                try:
                    metas[a] = int(metas[a])
                except:
                    metas[a] = float(metas[a])
        try:
            if metas.has_key('name'):
                try:
                    repository.rename(path, metas['name'])   
                except (OSError, IOError), e:
                    return web.notfound()
                del metas['name']
            if metas.has_key('symlink'):
                try:
                    repository.symlink(path, metas['symlink'])
                except (OSError, IOError), e:
                    return web.notfound()
                del metas['symlink']
            if metas.keys():
                try:
                    repository.update_metas(path, metas)
                except (OSError, IOError), e:
                    return web.notfound()
        except Exception, e:
            log.err("in POST %s:%s" % (e.__class__, e))
            web.internalerror()
        return 'done'

    def DELETE(self, repository_name, path):
        if repository_name not in repositories:
            return web.notfound()
        try:
            environ = web.webapi.ctx.environ
            repository = repositories[repository_name]
            repository.remove(path)
        except Exception, e:
            log.err("in DELETE %s:%s" % (e.__class__, e))
            return web.internalerror()

    def GET(self, repository_name, path=''):
        if repository_name not in repositories:
            return web.notfound()

        repository = repositories[repository_name]

        last_version = repository.last_version()
        if not last_version:
            web.internalerror()

        i = web.input(version=last_version)
        try:
            st = repository.stat(path)
        except OSError, e:
            log.err('got exc %s on stat(%s)' % (e, path,))
            return web.notfound()

        path = path.lstrip('/').rstrip('/')
        if stat.S_ISDIR(st.st_mode): 
            dir_list = repository.list(path=path, ver=int(i.version))
            vars_rendering = {} 
            vars_rendering['path'] = path
            vars_rendering['displaypath'] = cgi.escape(urllib.unquote(path))
            vars_rendering['repository_name'] = repository_name
            vars_rendering['files'] = []
            try:
                for  f in dir_list:
                    st = repository.stat('%s/%s' % (path, f))
                    displayname = f
                    type = 'file'
                    if stat.S_ISDIR(st.st_mode): 
                        displayname += '/'
                        type = 'directory'
                    linkname = "/repositories/%s/%s/%s" % (repository_name, path, f)
                    r = re.compile('/+')
                    linkname = r.sub('/', linkname)
                    vars_rendering['files'].append(
                                { 
                                'linkname' : urllib.quote(linkname), 
                                'displayname': cgi.escape(displayname), 
                                'name': f,
                                'mtime':  datetime.fromtimestamp(st.st_mtime).strftime('%Y/%m/%d %H:%M:%S'),
                                'size' : str(st.st_size),
                                'type': type,
                                }
                            )
            except (OSError, IOError), e:
                log.err("in GET %s:%s" % (e.__class__, e))
                return web.notfound()
            except Exception, e:
                log.err("in GET %s:%s" % (e.__class__, e))
                return web.internalerror()
            return data_rendering(vars_rendering) 
        elif stat.S_ISREG(st.st_mode):
            try:
                fd = repository.get_fd(path)
                data = fd.read()
                file_length = len(data)
                filename = os.path.basename(path)
                content_type = guess_type(path)[0] or 'text/plain'
                web.header("Content-type", content_type)
                web.header("Content-Length", file_length)
                web.header("Content-Disposition", 
                            'attachment; filename="%s" '% filename)
                return data
            except Exception, e:
                log.err("in GET: %s:%s" % (e.__class__, e))
                web.internalerror()
        return 'done'

@mimerender(
    default = 'txt',
    txt  = render_txt,
    html = render_html,
#    xml  = render_xml,
#   json = render_json,
    override_input_key='format',
)
def data_rendering(vars_rendering):
    return vars_rendering

#File Line iteration for Streaming 
def iter_fd(fd):
    buffer = []
    buffer_size = 4096
    while 1:
        if len(buffer) > 1:
            yield buffer.pop(0)
            continue
        chunks = fd.read(buffer_size).splitlines(True)
        first_chunk = buffer and buffer[0] or ''
        if chunks:
            first_chunk += chunks.pop(0)
        buffer = chunks
        if not first_chunk:
            return
        yield first_chunk

#Auth Middleware
def auth_mw(app):
    class AuthMiddleware(object):
        def __init__(self, application, realm, users):
            self.application = application
            self.realm = realm
            self.users = users
        def __call__(self, environ, start_response):
            path = environ['PATH_INFO']
            path = path.lstrip('/').rstrip('/')
            repository = ''
            if path.startswith('repositories'):
                repository =  path.split('/', 2)[1]
            else: 
                return self.application(environ, start_response)
            log.debug('checking HTTP_AUTHORIZATION for repository %s' % repository)
            if repository in repositories and 'HTTP_AUTHORIZATION' in environ:
                hdr = environ['HTTP_AUTHORIZATION']
                username, password = b64decode(hdr[6:].strip()).split(':')
                if username == self.users[repository]['login'] and\
                    password == self.users[repository]['password']:
                    return self.application(environ, start_response)
            status = '401 Unauthorized'
            headers = [('WWW-Authenticate', 'Basic realm="%s"' % self.realm)]
            start_response(status, headers)
            return 'Unauthorized\n'

    return AuthMiddleware(app, web.websafe(realm), accesses)

#middleware helper
class middleware:
    """Helper to convert wsgi middleware to application processor.
    Usage:
    app.add_processor(middleware(mw))
    """
    count = 0
    def __init__(self, middleware):
        self.name = self._get_name()
        self.wsgi = middleware(self.fake_wsgi_app)

    def _get_name(self):
        middleware.count += 1
        return "_wsgi_middleware_handler_%d" % (middleware.count)

    def __call__(self, handler):
        web.ctx[self.name] = handler
        environ = web.ctx.environ
        def start_response(status, headers):
            web.ctx.status = status
            web.ctx.headers = headers
        return self.wsgi(environ, start_response)

    def fake_wsgi_app(self, environ, start_response):
        handler = web.ctx[self.name]
        try:
            out = handler()
        except web.webapi.HTTPError, e:
            out = e.data
        start_response(web.ctx.status, web.ctx.headers)
        return out

def notfound():
    return web.notfound("Sorry, the page you were looking for was not found.")

def internalerror():
    return web.internalerror("InternalError")

def defaultdeny():
    print "Denied access"
    return "Access denied."

